@@ -60,11 +60,6 @@ enum mailbox_flags {
 	   quota updates (possibly resulting in broken quota). and This is
 	   useful for example when deleting entire user accounts. */
 	MAILBOX_FLAG_DELETE_UNSAFE	= 0x400,
-	/* Mailbox is used for caching purposes. Some of the mails may be
-	   stubs, which exist in the index but that don't have a mail body.
-	   The backend shouldn't treat it as corruption if a mail body isn't
-	   found. */
-	MAILBOX_FLAG_USE_STUBS		= 0x800,
 	/* Mailbox is created implicitly if it does not exist. */
 	MAILBOX_FLAG_AUTO_CREATE	= 0x1000,
 	/* Mailbox is subscribed to implicitly when it is created automatically */
@@ -73,6 +68,13 @@ enum mailbox_flags {
 	   useful in fixing index corruption errors that aren't otherwise
 	   detected and that are causing the full mailbox opening to fail. */
 	MAILBOX_FLAG_FSCK		= 0x4000,
+	/* Interpret name argument for mailbox_alloc_for_user() as a SPECIAL-USE
+	   flag. */
+	MAILBOX_FLAG_SPECIAL_USE	= 0x8000,
+	/* Mailbox is opened for reading/writing attributes. This allows ACL
+	   plugin to determine correctly whether the mailbox should be allowed
+	   to be opened. */
+	MAILBOX_FLAG_ATTRIBUTE_SESSION	= 0x10000,
 };
 
 enum mailbox_feature {
@@ -180,7 +182,8 @@ enum mail_fetch_field {
 	MAIL_FETCH_GUID			= 0x00200000,
 	MAIL_FETCH_POP3_ORDER		= 0x00400000,
 	MAIL_FETCH_REFCOUNT		= 0x00800000,
-	MAIL_FETCH_BODY_SNIPPET		= 0x01000000
+	MAIL_FETCH_BODY_SNIPPET		= 0x01000000,
+	MAIL_FETCH_REFCOUNT_ID		= 0x02000000,
 };
 
 enum mailbox_transaction_flags {
@@ -204,10 +207,6 @@ enum mailbox_transaction_flags {
 	   especially means the notify plugin. This would normally be used only
 	   with _FLAG_SYNC. */
 	MAILBOX_TRANSACTION_FLAG_NO_NOTIFY	= 0x40,
-	/* Append fills in an existing stub mail for the specified UID,
-	   instead of saving a new mail. This requires mailbox to be opened
-	   with MAILBOX_FLAG_USE_STUBS. */
-	MAILBOX_TRANSACTION_FLAG_FILL_IN_STUB	= 0x80,
 };
 
 enum mailbox_sync_flags {
@@ -379,7 +378,12 @@ enum mail_lookup_abort {
 	   operations to be performed, such as stat()ing a file. */
 	MAIL_LOOKUP_ABORT_READ_MAIL,
 	/* Abort if the operation can't be done fully using cache file */
-	MAIL_LOOKUP_ABORT_NOT_IN_CACHE
+	MAIL_LOOKUP_ABORT_NOT_IN_CACHE,
+	/* Abort if the operation can't be done fully using cache file.
+	 * During this lookup all cache lookups that have "no" decision
+	 * will be changed to "tmp". This way the field will start to be
+	 * cached in the future. */
+	MAIL_LOOKUP_ABORT_NOT_IN_CACHE_START_CACHING,
 };
 
 enum mail_access_type {
@@ -394,7 +398,6 @@ struct mail {
 	/* always set */
 	struct mailbox *box;
 	struct mailbox_transaction_context *transaction;
-	struct event *event;
 	uint32_t seq, uid;
 
 	bool expunged:1;
@@ -402,9 +405,10 @@ struct mail {
 	bool has_nuls:1; /* message data is known to contain NULs */
 	bool has_no_nuls:1; /* -''- known to not contain NULs */
 
-	/* Mail's header/body stream was opened within this request.
-	   If lookup_abort!=MAIL_LOOKUP_ABORT_NEVER, this can't become TRUE. */
-	bool mail_stream_opened:1;
+	/* Mail's header/body stream was opened (or attempted to be opened)
+	   within this request. If lookup_abort!=MAIL_LOOKUP_ABORT_NEVER, this
+	   can't become TRUE. */
+	bool mail_stream_accessed:1;
 	/* Mail's fast metadata was accessed within this request, e.g. the mail
 	   file was stat()ed. If mail_stream_opened==TRUE, this value isn't
 	   accurate anymore, because some backends may always set this when
@@ -517,13 +521,14 @@ struct mailbox *mailbox_alloc(struct mailbox_list *list, const char *vname,
 struct mailbox *mailbox_alloc_guid(struct mailbox_list *list,
 				   const guid_128_t guid,
 				   enum mailbox_flags flags);
-/* Initialize mailbox for delivery without actually opening any files or
-   verifying that it exists. */
-struct mailbox *mailbox_alloc_delivery(struct mail_user *user,
-	const char *name, enum mailbox_flags flags);
-/* Set a human-readable reason for why this mailbox is being accessed.
-   This is used for logging purposes. */
-void mailbox_set_reason(struct mailbox *box, const char *reason);
+/* Initialize mailbox for a particular user without actually opening any files
+   or verifying that it exists. The mname parameter is normally equal to the
+   mailbox vname, except when the MAILBOX_FLAG_SPECIAL_USE flag is set, in which
+   case it is the special-use flag. */
+struct mailbox *
+mailbox_alloc_for_user(struct mail_user *user, const char *mname,
+		       enum mailbox_flags flags);
+
 /* Get mailbox existence state. If auto_boxes=FALSE, return
    MAILBOX_EXISTENCE_NONE for autocreated mailboxes that haven't been
    physically created yet */
@@ -546,6 +551,9 @@ bool mailbox_equals(const struct mailbox *box1,
 /* Returns TRUE if the mailbox is user's INBOX or another user's shared INBOX */
 bool mailbox_is_any_inbox(struct mailbox *box);
 
+/* Returns TRUE if the mailbox has the specified special use flag assigned. */
+bool mailbox_has_special_use(struct mailbox *box, const char *special_use);
+
 /* Change mailbox_verify_create_name() to not verify new mailbox name
    restrictions (but still check that it's a valid existing name). This is
    mainly used by dsync to make sure the sync works even though the original
@@ -728,6 +736,10 @@ bool mailbox_search_next_nonblock(struct mail_search_context *ctx,
    determine correctly if those messages should have been returned in this
    search. */
 bool mailbox_search_seen_lost_data(struct mail_search_context *ctx);
+/* Detach the given mail from the search context. This allows the mail to live
+   even after mail_search_context has been freed. */
+void mailbox_search_mail_detach(struct mail_search_context *ctx,
+				struct mail *mail);
 
 /* Remember the search result for future use. This must be called before the
    first mailbox_search_next*() call. */
@@ -763,6 +775,10 @@ mailbox_keywords_create_valid(struct mailbox *box,
 struct mail_keywords *
 mailbox_keywords_create_from_indexes(struct mailbox *box,
 				     const ARRAY_TYPE(keyword_indexes) *idx);
+/* Return union of two mail_keywords. They must be created in the same
+   mailbox. */
+struct mail_keywords *mailbox_keywords_merge(struct mail_keywords *keywords1,
+					     struct mail_keywords *keywords2);
 void mailbox_keywords_ref(struct mail_keywords *keywords);
 void mailbox_keywords_unref(struct mail_keywords **keywords);
 /* Returns TRUE if keyword is valid, FALSE and error if not. */
@@ -855,6 +871,9 @@ void mail_add_temp_wanted_fields(struct mail *mail,
 				 struct mailbox_header_lookup_ctx *headers)
 	ATTR_NULL(3);
 
+/* Returns mail's event. This lazily creates the event when called for the
+   first time for the mail. */
+struct event *mail_event(struct mail *mail);
 /* Returns message's flags */
 enum mail_flags mail_get_flags(struct mail *mail);
 /* Returns message's keywords */
@@ -876,8 +895,9 @@ int mail_get_parts(struct mail *mail, struct message_part **parts_r);
 int mail_get_date(struct mail *mail, time_t *date_r, int *timezone_r);
 /* Get the time when the mail was received (IMAP INTERNALDATE). */
 int mail_get_received_date(struct mail *mail, time_t *date_r);
-/* Get the time when the mail was saved into this mailbox. This time may not
-   always be entirely reliable. */
+/* Get the time when the mail was saved into this mailbox. This returns -1 on
+   error, 0 if a real save date is not supported and a fall-back date is used,
+   and 1 when a save date was successfully retrieved. */
 int mail_get_save_date(struct mail *mail, time_t *date_r);
 
 /* Get the space used by the mail as seen by the reader. Linefeeds are always
@@ -939,7 +959,8 @@ int mail_get_hdr_stream_because(struct mail *mail,
 /* Returns the message part's body decoded to 8bit binary. If the
    Content-Transfer-Encoding isn't supported, returns -1 and sets error to
    MAIL_ERROR_CONVERSION. If the part refers to a multipart, all of its
-   children are returned decoded. */
+   children are returned decoded. Note that the returned stream must be
+   unreferenced, unlike mail_get_*stream*() which automatically free it. */
 int mail_get_binary_stream(struct mail *mail, const struct message_part *part,
 			   bool include_hdr, uoff_t *size_r,
 			   bool *binary_r, struct istream **stream_r);
@@ -955,6 +976,14 @@ int mail_get_special(struct mail *mail, enum mail_fetch_field field,
    but in virtual mailboxes it points to the backend mailbox. */
 int mail_get_backend_mail(struct mail *mail, struct mail **real_mail_r);
 
+/* Retrieve and parse the value of the Message-ID header field. Returns 1 if the
+   header was found and it contains a valid message ID, 0 if the header was not
+   found or no valid message ID was contained in it, and -1 if an error occurred
+   while retrieving the header. Returns the message ID value including '<' and
+   '>' in the *value_r return parameter or NULL if the header wasn't found or
+   its value was invalid. */
+int mail_get_message_id(struct mail *mail, const char **value_r);
+
 /* Update message flags. */
 void mail_update_flags(struct mail *mail, enum modify_type modify_type,
 		       enum mail_flags flags);
@@ -972,7 +1001,7 @@ void mail_update_pop3_uidl(struct mail *mail, const char *uidl);
 void mail_expunge(struct mail *mail);
 
 /* Add missing fields to cache. */
-void mail_precache(struct mail *mail);
+int mail_precache(struct mail *mail);
 /* Mark a cached field corrupted and have it recalculated. */
 void mail_set_cache_corrupted(struct mail *mail,
 			      enum mail_fetch_field field,
